#include <arpa/inet.h>
#include <errno.h>
#include <fcntl.h>
#include <grp.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <syslog.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <unistd.h>

#include <config.h>

#ifdef HAVE_SYS_CAPABILITY_H
#include <sys/capability.h>
#endif

#ifdef HAVE_TCP_WRAPPERS
#include <tcpd.h>
#endif

#include "bytecode.h"
#include "db.h"
#include "rlinetd.h"
#include "signals.h"

/* we avoid the more tedious sanity checks by assuming that since
	 we wrote the opcodes, they're going to be what we accept. dodgy. */

#define rl_fail1(a) \
  rl_warn("%s(%d) failed %s", (a), *op, strerror(errno))

#define RL_MIN(a, b) ((a) < (b) ? (a) : (b))

int allow_severity = LOG_INFO;
int deny_severity = LOG_WARNING;

void run_bytecode(rl_opcode_t *op, struct rl_instance *inst) {
	if(!op)
		return;
	do {
		if(rl_debug > 1)
			fprintf(stderr, "op %d\n", *op);
		switch(*op) {
			struct argvtab *argv;
			struct logtab *log;
			int i, j, k;
#ifdef HAVE_TCP_WRAPPERS
			struct request_info *request;
#endif			
			char *str;
			struct maptab *mt;
			gid_t gid;
			struct semaphore *sem;
			
			case OP_EXEC:
				closelog();
				for(i = 0; i < 3; i++)
					if(inst->sock != i)
						dup2(inst->sock, i);
				if(inst->sock >= 3)
					close(inst->sock);
				j = *++op;
				argv = argvtab_get(*++op);
				argv_build(inst, argv);
				str = (j != -1) ? stringtab_get(j) : argv->argv[0];
				if(execve(str, argv->argv, NULL)) {
					rl_warn("execve(%s): %s", str, strerror(errno));
					exit(EX_DATAERR);
				}
				break;
			case OP_FISH:
				rl_warn("Fish!");
				break;
			case OP_SUID:
				if(setuid(*++op))
					rl_fail1("setuid");
				break;
			case OP_SGID:
				if(setgid(*++op))
					rl_fail1("setgid");
				break;
			case OP_NICE:
				if(setpriority(PRIO_PROCESS, 0, *++op))
					rl_fail1("setpriority");
				break;
			case OP_RLIMIT:
				if(setrlimit(*++op, rlimittab_get(*++op)))
					rl_fail1("setrlimit");
				break;
			case OP_CHROOT:
				argv = argvtab_get(*++op);
				string_build(inst, argv);
				if(chroot(argv->str))
					rl_warn("chroot(\"%s\"): %s", argv->str, strerror(errno));
				if(chdir("/"))
					rl_warn("chdir(\"/\"): %s", strerror(errno));
				break;
			case OP_LOG:
				log = logtab_get(*++op);
				argv = argvtab_get(log->argv);
				if(log->fd < 0) {
					string_build(inst, argv);
					syslog(LOG_INFO, argv->str);
				} else {
					iov_build(inst, argv);
					if(writev(log->fd, argv->iov, argv->argc) < 0)
						rl_warn("writev() log: %s", strerror(errno));
				}
				break;
			case OP_ACCEPT:
				i = *++op;

				if(inst->sin)
					free(inst->sin);
				inst->sin = NULL;
				inst->sinlen = i;
				if(i)
					inst->sin = (struct sockaddr *)malloc(i);
				if((inst->sock = accept(inst->sock, inst->sin, &inst->sinlen)) < 0) {
					rl_warn("accept(): %s",	strerror(errno));
					return;
				}
				if(fcntl(inst->sock, F_SETFD, 0) == -1) {
					rl_warn("fcntl(%d, F_SETFD, 0): %s", inst->sock, strerror(errno));
				}
				break;
			case OP_FORK:
				k = *++op;
				i = *++op;
				rls_block();
				switch(j = fork()) {
					case -1:
						rl_warn("fork() failed - %s", strerror(errno));
						break;
					case 0:
						rls_unblock();
						all_unhook();
						inst->reg = 0;
						run_bytecode(++op, inst);
						exit(inst->reg);
						break;
					default:
						inst->start = time(NULL);
						pidtab_add(j, i, inst);
						rls_unblock();
						return run_bytecode(oplisttab_get(k), inst);
				}
				break;
			case OP_WRAP:
				str = stringtab_get(*++op);
				k = *++op;
#ifdef HAVE_TCP_WRAPPERS
				request = (struct request_info *)malloc(sizeof(*request));
				request_init(request,
										 RQ_FILE, inst->sock,
										 RQ_CLIENT_SIN, inst->sin,
										 RQ_DAEMON, str,
										 0);
				fromhost(request);			/* undocumented tcp wrappers magic -
																does DNS lookup and possibly other
																stuff */
				if(!hosts_access(request)) {
					run_bytecode(oplisttab_get(k), inst);
				}
#endif				
				break;
			case OP_SETCAP:
#ifdef HAVE_LIBCAP
				if(cap_set_proc(captab_get(*++op)))
					rl_warn("cap_set_proc(): %s",	strerror(errno));
#endif				
				break;
			case OP_INITGR:
				str = stringtab_get(*++op);
				gid = (gid_t)*++op;
				if(initgroups(str, gid))
					rl_warn("initgroups(%s, %d): %s", str, gid, strerror(errno));
				break;
			case OP_BRANCH:
				j = *++op;
				return run_bytecode(oplisttab_get(j), inst);
			case OP_CLOSE:
				if(close(inst->sock))
					rl_warn("close(%d): %s", inst->sock, strerror(errno));
				inst->sock = -1;
				return;
			case OP_BANNER:
				mt = maptab_get(*++op);

				j = write(inst->sock, mt->addr + inst->reg,
									RL_MIN(mt->len - inst->reg, 4096));
				if(j < 0) {
					switch(errno) {
						case EAGAIN:
							j = 0;
							break;
						default:
							rl_warn("write() failed (%s)", strerror(errno));
							write_unhook(j);
							return;
					}
				}
				inst->reg += j;
				if(inst->reg < mt->len) {
					write_hook(inst->sock, --op, inst);
					return;
				}
				write_unhook(j);
				break;
			case OP_ZERO:
				inst->reg = 0;
				break;
			case OP_RET:
				return;
			case OP_ECHO:
				argv = argvtab_get(*++op);
				
				iov_build(inst, argv);
				if(writev(inst->sock, argv->iov, argv->argc) < 0)
					rl_warn("writev(): %s",	strerror(errno));
				break;
			case OP_UP:
				sem = semaphore_get(*++op);
				if(sem->count-- == sem->limit)
					run_bytecode(oplisttab_get(sem->under), inst);
				break;
			case OP_DOWN:
				sem = semaphore_get(*++op);
				if(++sem->count == sem->limit)
					run_bytecode(oplisttab_get(sem->match), inst);
				break;
			case OP_FROG:
				rl_warn("Frog!");
				break;
			case OP_LSET:
				listeners_set(*++op);
				break;
			case OP_LCLR:
				listeners_clear(*++op);
				break;
			default:
																/* should never happen */
				rl_fatal(EX_SOFTWARE, "ABORT - Unknown opcode (%d)", *op);
				break;
		}
	} while(op++);
}
